"""
Managers for our models
"""

from django.contrib.auth.models import User
from django.db import models
from django.db.models import Prefetch, Q


class SettingsManager(models.Manager):
    """
    SettingsManager
    """

    def get_setting(self, setting_key: str) -> str:
        """
        Return the value for given setting key
        """

        return self.get(variable=setting_key).value


class BoardQuerySet(models.QuerySet):
    def user_has_access(self, user: User) -> models.QuerySet:
        """
        Filter boards that given user has access to.
        """

        return self.filter(
            Q(groups__in=user.groups.all()) | Q(groups__isnull=True)
        ).distinct()


class BoardManagerBase(models.Manager):
    """
    BoardManagerBase
    """

    pass


BoardManager = BoardManagerBase.from_queryset(BoardQuerySet)


class TopicQuerySet(models.QuerySet):
    def get_from_slugs(
        self,
        category_slug: str,
        board_slug: str,
        topic_slug: str,
        user: User,
    ) -> models.Model:
        """
        Fetch topic from slugs for user. Return None if not found or no access.
        """

        from .models import Message

        try:
            topic = (
                self.select_related(
                    "board",
                    "board__category",
                    "first_message",
                    "first_message__topic",
                    "first_message__topic__board",
                    "first_message__topic__board__category",
                )
                .prefetch_related(
                    Prefetch(
                        "messages",
                        queryset=Message.objects.select_related(
                            "user_created", "user_created__profile__main_character"
                        ).order_by("time_posted"),
                        to_attr="messages_sorted",
                    )
                )
                .filter(
                    board__category__slug=str(category_slug),
                    board__slug=str(board_slug),
                    slug=str(topic_slug),
                )
                .filter(
                    Q(board__groups__in=user.groups.all())
                    | Q(board__groups__isnull=True)
                )
                .distinct()
                .get()
            )
        except self.model.DoesNotExist:
            return None

        return topic


class TopicManagerBase(models.Manager):
    """
    TopicManagerBase
    """

    pass


TopicManager = TopicManagerBase.from_queryset(TopicQuerySet)



from django.db import migrations

default_settings_to_migrate = [
    {"variable": "defaultMaxMessages", "value": "15"},
    {"variable": "defaultMaxTopics", "value": "20"},
]


def add_default_setting(apps, schema_editor):
    """
    Add default settings on migration
    :param apps:
    :type apps:
    :param schema_editor:
    :type schema_editor:
    """

    Setting = apps.get_model("aa_forum", "Setting")
    db_alias = schema_editor.connection.alias
    default_settings = [
        Setting(variable=default_setting["variable"], value=default_setting["value"])
        for default_setting in default_settings_to_migrate
    ]
    Setting.objects.using(db_alias).bulk_create(default_settings)


def remove_default_settings(apps, schema_editor):
    """
    Remove default settings on migration to zero
    :param apps:
    :type apps:
    :param schema_editor:
    :type schema_editor:
    """

    Setting = apps.get_model("aa_forum", "Setting")
    db_alias = schema_editor.connection.alias

    for default_setting in default_settings_to_migrate:
        Setting.objects.using(db_alias).filter(
            variable=default_setting["variable"]
        ).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("aa_forum", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(add_default_setting, remove_default_settings),
    ]



import json

from django.test import TestCase
from django.urls import reverse

from ..models import Board, Category
from .utils import create_fake_user

VIEWS_PATH = "aa_forum.views.admin"


class TestAdminViews(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = create_fake_user(
            1001,
            "Bruce Wayne",
            permissions=["aa_forum.basic_access", "aa_forum.manage_forum"],
        )

    def test_should_delete_category(self):
        # given
        category = Category.objects.create(name="Category")
        self.client.force_login(self.user)
        # when
        # when
        res = self.client.get(
            reverse("aa_forum:admin_category_delete", args=[category.pk])
        )
        # then
        self.assertRedirects(res, reverse("aa_forum:admin_index"))
        self.assertFalse(Category.objects.filter(pk=category.pk).exists())

    def test_should_raise_404_when_delete_category_not_found(self):
        # given
        Category.objects.create(name="Category")
        self.client.force_login(self.user)
        # when
        res = self.client.get(reverse("aa_forum:admin_category_delete", args=[0]))
        # then
        self.assertEqual(res.status_code, 404)

    def test_should_delete_board(self):
        # given
        category = Category.objects.create(name="Category")
        board = Board.objects.create(name="Board", category=category)
        self.client.force_login(self.user)
        # when
        # when
        res = self.client.get(
            reverse("aa_forum:admin_board_delete", args=[category.pk, board.pk])
        )
        # then
        self.assertRedirects(res, reverse("aa_forum:admin_index"))
        self.assertFalse(Board.objects.filter(pk=board.pk).exists())

    def test_should_raise_404_when_delete_board_not_found(self):
        # given
        category = Category.objects.create(name="Category")
        Board.objects.create(name="Board", category=category)
        self.client.force_login(self.user)
        # when
        res = self.client.get(
            reverse("aa_forum:admin_board_delete", args=[category.pk, 0])
        )
        # then
        self.assertEqual(res.status_code, 404)

    def test_should_save_categories_order(self):
        # given
        category_1 = Category.objects.create(name="Category 1")
        category_2 = Category.objects.create(name="Category 2")
        self.client.force_login(self.user)
        # when
        res = self.client.post(
            reverse("aa_forum:admin_ajax_category_order"),
            data={
                "categories": json.dumps(
                    [
                        {"catId": category_1.pk, "catOrder": 1},
                        {"catId": category_2.pk, "catOrder": 2},
                    ]
                )
            },
        )
        # then
        self.assertListEqual(res.json(), [{"success": True}])
        category_1.refresh_from_db()
        self.assertEqual(category_1.order, 1)
        category_2.refresh_from_db()
        self.assertEqual(category_2.order, 2)

    def test_should_save_categories_order_and_handle_errors(self):
        # given
        category_1 = Category.objects.create(name="Category 1")
        self.client.force_login(self.user)
        # when
        res = self.client.post(
            reverse("aa_forum:admin_ajax_category_order"),
            data={
                "categories": json.dumps(
                    [
                        {"catId": category_1.pk, "catOrder": 1},
                        {"catId": 0, "catOrder": 2},
                    ]
                )
            },
        )
        # then
        self.assertListEqual(res.json(), [{"success": True}])
        category_1.refresh_from_db()
        self.assertEqual(category_1.order, 1)

    def test_should_save_boards_order(self):
        # given
        category = Category.objects.create(name="Category")
        board_1 = Board.objects.create(name="Board 1", category=category)
        board_2 = Board.objects.create(name="Board 2", category=category)
        self.client.force_login(self.user)
        # when
        res = self.client.post(
            reverse("aa_forum:admin_ajax_board_order"),
            data={
                "boards": json.dumps(
                    [
                        {"boardId": board_1.pk, "boardOrder": 1},
                        {"boardId": board_2.pk, "boardOrder": 2},
                    ]
                )
            },
        )
        # then
        self.assertListEqual(res.json(), [{"success": True}])
        board_1.refresh_from_db()
        self.assertEqual(board_1.order, 1)
        board_2.refresh_from_db()
        self.assertEqual(board_2.order, 2)

    def test_should_save_boards_order_and_handle_errors(self):
        # given
        category = Category.objects.create(name="Category")
        board_1 = Board.objects.create(name="Board 1", category=category)
        self.client.force_login(self.user)
        # when
        res = self.client.post(
            reverse("aa_forum:admin_ajax_board_order"),
            data={
                "boards": json.dumps(
                    [
                        {"boardId": board_1.pk, "boardOrder": 1},
                        {"boardId": 0, "boardOrder": 2},
                    ]
                )
            },
        )
        # then
        self.assertListEqual(res.json(), [{"success": True}])
        board_1.refresh_from_db()
        self.assertEqual(board_1.order, 1)



import datetime as dt
import random
import re
from typing import List
from unittest.mock import patch

from faker import Faker

from django.contrib.auth.models import User
from django.utils.timezone import now

from allianceauth.tests.auth_utils import AuthUtils

from aa_forum.constants import SETTING_MESSAGESPERPAGE

from ..models import Message, Setting, Topic

MAX_MESSAGE_HOURS_INTO_PAST = 1000
fake = Faker()


def create_fake_user(
    character_id: int,
    character_name: str,
    corporation_id: int = None,
    corporation_name: str = None,
    corporation_ticker: str = None,
    alliance_id: int = None,
    alliance_name: str = None,
    permissions: List[str] = None,
) -> User:
    """Create a fake user incl. main character and (optional) permissions."""
    username = re.sub(r"[^\w\d@\.\+-]", "_", character_name)
    user = AuthUtils.create_user(username)
    if not corporation_id:
        corporation_id = 2001
        corporation_name = "Wayne Technologies Inc."
        corporation_ticker = "WTE"
    if not alliance_id:
        alliance_id = 3001
        alliance_name = "Wayne Enterprises"
    AuthUtils.add_main_character_2(
        user=user,
        name=character_name,
        character_id=character_id,
        corp_id=corporation_id,
        corp_name=corporation_name,
        corp_ticker=corporation_ticker,
        alliance_id=alliance_id,
        alliance_name=alliance_name,
    )
    if permissions:
        perm_objs = [AuthUtils.get_permission_by_name(perm) for perm in permissions]
        user = AuthUtils.add_permissions_to_user(perms=perm_objs, user=user)
    return user


def create_fake_message(topic: Topic, user: User):
    """Create a fake message."""
    return Message.objects.create(
        topic=topic, message=f"<p>{fake.sentence()}</p>", user_created=user
    )


def create_fake_messages(topic: Topic, amount) -> List[Message]:
    """Create a bunch of fake messags in given topic."""
    users = list(User.objects.all())
    messages = list()
    with patch("django.utils.timezone.now", new=random_dt):
        for _ in range(amount):
            messages.append(create_fake_message(topic, user=random.choice(users)))
    return messages


def random_dt() -> dt.datetime:
    """Return random datetime between now and x hours into the past."""
    return now() - dt.timedelta(
        hours=random.randint(0, MAX_MESSAGE_HOURS_INTO_PAST),
        minutes=random.randint(0, 59),
        seconds=random.randint(0, 59),
    )


def my_get_setting(setting_key: str) -> str:
    """Overload settings for tests."""
    if setting_key == SETTING_MESSAGESPERPAGE:
        return "5"
    return Setting.objects.get_setting(setting_key=setting_key)
